"""JSON Blob Cache"""
import functools
import itertools
import json
import operator
from pathlib import Path
from types import SimpleNamespace
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)
from uuid import UUID

import lmdb
import toolz
import zstandard

# pylint: disable=invalid-name,too-many-public-methods

FORMAT_VERSION = b"0001"
RESERVED_INFO_KEYS = {"version", "zdict"}

Prefix = Union[str, bytes]
LaxUUID = Union[str, UUID]
Document = Dict[str, Any]
Document_ = Union[Document, Tuple[Union[str, UUID], Document]]
MaybeTransaction = Optional[lmdb.Transaction]


def key_to_bytes(k: Any) -> bytes:
    if isinstance(k, str):
        return k.encode("utf8")
    if isinstance(k, UUID):
        return k.bytes
    if isinstance(k, bytes):
        return k
    if isinstance(k, int):
        if k.bit_length() < 32:
            return k.to_bytes(4, "big")
        if k.bit_length() < 128:
            return k.to_bytes(16, "big")
        return str(k).encode("utf8")
    if isinstance(k, tuple):
        return functools.reduce(operator.add, map(key_to_bytes, k))

    raise ValueError("Key must be one of str|bytes|int|UUID|tuple")


def uuids2bytes(uu: Collection[UUID]) -> bytes:
    bb = bytearray(len(uu) * 16)
    for i, u in enumerate(uu):
        bb[i * 16 : (i + 1) * 16] = u.bytes
    return bytes(bb)


def bytes2uuids(bb: bytes) -> List[UUID]:
    n = len(bb) // 16
    return [UUID(bytes=bb[i * 16 : (i + 1) * 16]) for i in range(n)]


def prefix_visit(
    tr: lmdb.Transaction, prefix: Prefix, full_key: bool = False, db=None
) -> Iterator[Tuple[bytes, bytes]]:
    if isinstance(prefix, str):
        prefix = prefix.encode("utf8")

    n = len(prefix)
    cursor = tr.cursor(db)
    cursor.set_range(prefix)
    for k, v in cursor:
        if len(k) < n or k[:n] != prefix:
            break
        yield (k if full_key else k[n:]), v


def dict2jsonKV(
    oo: Document,
    prefix: Optional[str] = None,
    compressor: Optional[zstandard.ZstdCompressor] = None,
) -> Iterator[Tuple[bytes, bytes]]:
    for k, doc in oo.items():
        data = json.dumps(doc, separators=(",", ":")).encode("utf8")
        if compressor is not None:
            data = compressor.compress(data)
        if prefix is not None:
            k = prefix + k

        yield (key_to_bytes(k), data)


def jsonKV2dict(
    kv: Iterable[Tuple[bytes, bytes]],
    decompressor: Optional[zstandard.ZstdDecompressor] = None,
) -> Document:
    def decode(kv):
        k, doc = kv
        if decompressor is not None:
            doc = decompressor.decompress(doc)
        return k.decode("utf8"), json.loads(doc)

    doc = dict(map(decode, kv))
    if len(doc) == 1 and list(doc)[0] == "":
        doc = doc[""]
    return doc


def doc2bytes(doc: Document_, purge_id: bool = False) -> Tuple[bytes, bytes]:
    """doc is

    Either:

    1. (id, {..}) key-value tuple
    2. {id: , ..}  dict with `id` field

    id has to be a UUID (typically string)
    """
    if isinstance(doc, tuple):
        k, d = doc
    else:
        k_ = doc.get("id")
        if not isinstance(k_, (str, UUID)):
            raise ValueError("Expect id key to be a string or a UUID")

        k = k_
        d = toolz.dissoc(doc, ["id"]) if purge_id else doc

    if not isinstance(k, UUID):
        if not isinstance(k, str):
            raise ValueError("Expect id key to be a string or a UUID")

        k = UUID(k)

    raw_k = k.bytes
    raw_doc = json.dumps(d, separators=(",", ":")).encode("utf8")
    return (raw_k, raw_doc)


class JsonBlobCache:
    """
    info:
       version: 4-bytes
       zdict: pre-trained compression dictionary, optional

    groups:
       Each group is named list of uuids

    udata:
       arbitrary user data (TODO)

    ds:
       uuid: compressed(json({..}))
    """

    def __init__(self, state):
        """Don't use this directly, use create_cache or open_(rw|ro)."""

        self._dbs = state.dbs
        self._comp: Optional[zstandard.ZstdCompressor] = state.comp
        self._decomp: zstandard.ZstdDecompressor = state.decomp
        self._closed = False
        self._current_transaction: MaybeTransaction = None

    def close(self):
        """Write any outstanding product/metadata definitions (if in write mode) and
        close database file.

        """
        if not self._closed:
            try:
                self._dbs.main.close()
            finally:
                self._closed = True

    def __del__(self):
        self.close()

    @staticmethod
    def train_dictionary(
        docs: Iterable[Document_], dict_sz: int = 8 * 1024
    ) -> Optional[bytes]:
        """Given a finite sequence of Documents train zstandard compression dictionary of a given size.

        Document is either a
         - (id, {..}) tuple
         - {id: str|UUID, ...} a dictionary with `id` key of type UUID (possibly string formatted UUID)

        Will return None if input sequence is empty.
        """
        sample = list(v for _, v in map(doc2bytes, docs))

        if len(sample) == 0:
            return None

        return zstandard.train_dictionary(dict_sz, sample).as_bytes()  # type: ignore

    def _append_info_dict(self, prefix: str, oo: Document, tr: lmdb.Transaction):
        for k, v in dict2jsonKV(oo, prefix, self._comp):
            tr.put(k, v, overwrite=True, dupdata=False, db=self._dbs.info)

    def _clear_info_dict(self, prefix: Prefix, tr: lmdb.Transaction):
        db_info = self._dbs.info
        for k, _ in prefix_visit(tr, prefix, full_key=True, db=db_info):
            tr.delete(k, db=db_info)

    def _get_info_dict(self, prefix: Prefix, tr: lmdb.Transaction) -> Document:
        return jsonKV2dict(prefix_visit(tr, prefix, db=self._dbs.info), self._decomp)

    def append_info_dict(
        self, prefix: str, oo: Document, transaction: MaybeTransaction = None
    ):
        if prefix in RESERVED_INFO_KEYS:
            raise ValueError(
                f"Prefix '{prefix}' can not be used, it is reserved for internal use"
            )

        if transaction is None:
            with self._dbs.main.begin(write=True) as tr:
                self._append_info_dict(prefix, oo, tr)
        else:
            self._append_info_dict(prefix, oo, transaction)

    def get_info_dict(
        self, prefix: str, transaction: MaybeTransaction = None
    ) -> Document:
        if prefix in RESERVED_INFO_KEYS:
            raise ValueError(
                f"Prefix '{prefix}' can not be read, it is reserved for internal use"
            )

        if transaction is None:
            with self._dbs.main.begin(write=False) as tr:
                return self._get_info_dict(prefix, tr)
        else:
            return self._get_info_dict(prefix, transaction)

    def clear_info_dict(self, prefix: str, transaction: MaybeTransaction = None):
        if prefix in RESERVED_INFO_KEYS:
            raise ValueError(
                f"Prefix '{prefix}' can not be cleared, it is reserved for internal use"
            )

        if transaction is None:
            with self._dbs.main.begin(write=True) as tr:
                return self._clear_info_dict(prefix, tr)
        else:
            return self._clear_info_dict(prefix, transaction)

    def get_info_keys(self, prefix: str = "") -> Iterator[str]:
        with self._dbs.main.begin(self._dbs.info, write=False, buffers=True) as tr:
            if prefix == "":
                kv_stream = tr.cursor()
            else:
                kv_stream = prefix_visit(tr, prefix, full_key=True)

            for k, _ in kv_stream:
                ks = bytes(k).decode("utf8")
                if ks not in RESERVED_INFO_KEYS:
                    yield ks

    @property
    def current_transaction(self) -> MaybeTransaction:
        return self._current_transaction

    @property
    def readonly(self) -> bool:
        return self._comp is None

    def _doc2kv(self, doc: Document_) -> Tuple[bytes, bytes]:
        k, d = doc2bytes(doc)
        if self._comp:
            d = self._comp.compress(d)
        return (k, d)

    def _doc_save(self, doc: Document_, transaction: lmdb.Transaction):
        k, v = self._doc2kv(doc)
        transaction.put(k, v)

    def bulk_save(self, docs: Iterable[Document_]):
        try:
            with self._dbs.main.begin(self._dbs.ds, write=True) as tr:
                self._current_transaction = tr
                for doc in docs:
                    self._doc_save(doc, tr)
        finally:
            self._current_transaction = None

    def tee(
        self, docs: Iterable[Any], max_transaction_size: int = 10000, transform=None
    ) -> Iterator[Any]:
        """Given a lazy stream of (k,v) pairs persist them to disk and then pass through
        for further processing.
        :docs: stream of documents (uuid, {..}) pairs
        :max_transaction_size int: How often to commit results to disk
        :transform: If docs contains some other objects you can supply mapping from that Data type to Tuple[uuid, dict]
        """
        have_some = True
        if transform is None:
            transform = lambda x: x

        try:
            while have_some:
                with self._dbs.main.begin(self._dbs.ds, write=True) as tr:
                    self._current_transaction = tr
                    have_some = False
                    for ds in itertools.islice(docs, max_transaction_size):
                        have_some = True
                        self._doc_save(transform(ds), tr)
                        yield ds
        finally:
            self._current_transaction = None

    def put_group(self, name: str, uuids: Collection[UUID]):
        """Group is a named list of uuids"""
        data = uuids2bytes(uuids)
        k = key_to_bytes(name)

        with self._dbs.main.begin(self._dbs.groups, write=True) as tr:
            tr.put(k, data)

    def _get_group_raw(self, name: Prefix) -> bytes:
        k = key_to_bytes(name)

        with self._dbs.main.begin(self._dbs.groups, write=False) as tr:
            return tr.get(k)

    def get_group(self, name: str) -> Optional[List[UUID]]:
        """Group is a named list of uuids"""
        data = self._get_group_raw(key_to_bytes(name))
        return bytes2uuids(data) if data is not None else None

    def groups(
        self, raw: bool = False, prefix: Optional[Prefix] = None
    ) -> Union[List[Tuple[bytes, int]], List[Tuple[str, int]]]:
        """Get list of tuples (group_name, group_size).

        :raw bool: Normally names are returned as strings, supplying raw=True
        would return bytes instead, this is needed if you are using group names
        that are not strings, like integers or tuples of basic types.

        :prefix str|bytes: Only report groups with name starting with prefix
        """

        assert isinstance(prefix, (str, bytes, type(None)))

        def _raw(prefix):
            with self._dbs.main.begin(
                self._dbs.groups, write=False, buffers=True
            ) as tr:
                cursor = (
                    tr.cursor()
                    if prefix is None
                    else prefix_visit(tr, prefix, full_key=True)
                )
                return [(bytes(k), len(d) // 16) for k, d in cursor]

        if prefix is not None:
            prefix = key_to_bytes(prefix)

        nn = _raw(prefix)
        return nn if raw else [(n.decode("utf8"), c) for n, c in nn]

    def _extract_ds(self, d: bytes) -> Document:
        d = self._decomp.decompress(d)
        return json.loads(d)

    def get(self, uuid: LaxUUID) -> Optional[Document]:
        """Extract single dataset with a given uuid, or return None if not found"""
        if isinstance(uuid, str):
            uuid = UUID(uuid)

        key = key_to_bytes(uuid)

        with self._dbs.main.begin(self._dbs.ds, buffers=True) as tr:
            d = tr.get(key, None)
            if d is None:
                return None

            return self._extract_ds(d)

    def get_all(self) -> Iterator[Tuple[UUID, Document]]:
        try:
            with self._dbs.main.begin(self._dbs.ds, buffers=True) as tr:
                self._current_transaction = tr
                for k, d in tr.cursor():
                    yield UUID(bytes=bytes(k)), self._extract_ds(d)
        finally:
            self._current_transaction = None

    def stream_group(self, group_name: str) -> Iterator[Tuple[UUID, Document]]:
        uu = self._get_group_raw(group_name)
        if uu is None:
            raise ValueError(f"No such group: {group_name}")

        if len(uu) & 0xF:
            raise ValueError(f"Wrong data size for group {group_name}")

        with self._dbs.main.begin(self._dbs.ds, buffers=True) as tr:
            for i in range(0, len(uu), 16):
                key = uu[i : i + 16]
                d = tr.get(key, None)
                if d is None:
                    raise ValueError(f"Missing dataset for {str(UUID(bytes=key))}")

                yield UUID(bytes=bytes(key)), self._extract_ds(d)

    @property
    def count(self) -> int:
        with self._dbs.main.begin(self._dbs.ds) as tr:
            return tr.stat()["entries"]

    @property
    def path(self) -> Path:
        return Path(self._dbs.main.path())

    @staticmethod
    def open_ro(path: str, lock: bool = False) -> "JsonBlobCache":
        return open_ro(path, lock=lock)

    @staticmethod
    def open_rw(path: str, **kw) -> "JsonBlobCache":
        return open_rw(path, **kw)

    @staticmethod
    def create(
        path: str,
        complevel: int = 6,
        zdict: Optional[bytes] = None,
        max_db_sz: Optional[int] = None,
        lock: bool = False,
        subdir: bool = False,
        truncate: bool = False,
        **kw,
    ) -> "JsonBlobCache":
        """Create new file database or open existing one.

        :path str: Path where to create new database (this will be a directory with 2 files in it)
        :complevel int: Level of compressions to apply per dataset, bigger is slower but better compression.
        :zdict: Optional pre-trained compression dictionary
        :max_db_sz int: Maximum size in bytes (defaults to 10GiB)
        :lock: By default we assume exclusive access to the file, if you expect sharing set ``lock=True``
        :subdir: If set to True treat path as directory

        :truncate bool: If True wipe out any existing database and create new empty one.
        :kw: Passed on to ``lmdb.open(.., **kw)``
           meminit
            writemap
            sync
            metasync
            ...and other
        """

        return create_cache(
            path,
            complevel=complevel,
            zdict=zdict,
            max_db_sz=max_db_sz,
            lock=lock,
            subdir=subdir,
            truncate=truncate,
            **kw,
        )

    @staticmethod
    def exists(path: str) -> bool:
        return db_exists(path)


def db_exists(path: str) -> bool:
    _path = Path(path)
    if not _path.exists():
        return False

    if _path.is_dir():
        return Path(_path / "data.mdb").exists()

    return True


def maybe_delete_db(path: str) -> bool:
    """Delete existing database if it exists.

     LMDB database consists of two files, data + lock, they can be arranged in two possible layouts:

     - `db-dir-name/{data.mdb, lock.mdb}`
     - `db-file-name` and `db-file-name-lock`

    You supply path which is either `db-dir-name` or `db-file-name`.
    """
    _path = Path(path)
    if not _path.exists():
        return False

    if _path.is_dir():
        db, lock = (_path / n for n in ["data.mdb", "lock.mdb"])
    else:
        db = _path
        lock = Path(str(_path) + "-lock")

    if db.exists() and lock.exists():
        db.unlink()
        lock.unlink()

        if _path.is_dir():
            _path.rmdir()

    return True


def _from_existing_db(db, complevel: int = 6) -> JsonBlobCache:
    readonly = db.flags().get("readonly")

    try:
        db_info = db.open_db(b"info", create=False)
    except lmdb.NotFoundError:
        raise ValueError("Existing database is not a ds cache") from None

    with db.begin(db_info, write=False) as tr:
        version = tr.get(b"version", None)
        if version is None:
            raise ValueError("Missing format version field")
        if version != FORMAT_VERSION:
            raise ValueError("Unsupported on disk version: " + version.decode("utf8"))

        zdict = tr.get(b"zdict", None)

    dbs = SimpleNamespace(
        main=db,
        info=db_info,
        groups=db.open_db(b"groups", create=False),
        ds=db.open_db(b"ds", create=False),
        udata=db.open_db(b"udata", create=False),
    )

    comp_params: Dict[str, Any] = (
        {"dict_data": zstandard.ZstdCompressionDict(zdict)} if zdict else {}
    )

    comp = (
        None if readonly else zstandard.ZstdCompressor(level=complevel, **comp_params)
    )
    decomp = zstandard.ZstdDecompressor(**comp_params)

    state = SimpleNamespace(dbs=dbs, comp=comp, decomp=decomp)

    return JsonBlobCache(state)


def _from_empty_db(db, complevel: int = 6, zdict: Optional[bytes] = None):
    assert isinstance(zdict, (bytes, type(None)))

    db_info = db.open_db(b"info", create=True)

    with db.begin(db_info, write=True) as tr:
        tr.put(b"version", FORMAT_VERSION)

        if zdict is not None:
            tr.put(b"zdict", zdict)

    dbs = SimpleNamespace(
        main=db,
        info=db_info,
        groups=db.open_db(b"groups", create=True),
        ds=db.open_db(b"ds", create=True),
        udata=db.open_db(b"udata", create=True),
    )

    comp_params: Dict[str, Any] = (
        {"dict_data": zstandard.ZstdCompressionDict(zdict)} if zdict else {}
    )

    comp = zstandard.ZstdCompressor(level=complevel, **comp_params)
    decomp = zstandard.ZstdDecompressor(**comp_params)

    state = SimpleNamespace(dbs=dbs, comp=comp, decomp=decomp)

    return JsonBlobCache(state)


def open_ro(path: str, lock: bool = False, **kw) -> JsonBlobCache:
    """Open existing database in readonly mode.

    .. note:

       default mode assumes db file is static (not being modified
       externally), if this is not the case, supply ``lock=True`` parameter.

    :path: Path to the db could be folder or actual file

    :lock: Supply ``True`` if external process is changing DB concurrently.
    """

    subdir = Path(path).is_dir()

    db = lmdb.open(
        path, subdir=subdir, max_dbs=8, lock=lock, create=False, readonly=True, **kw
    )

    return _from_existing_db(db)


def open_rw(
    path: str,
    max_db_sz: Optional[int] = None,
    complevel: int = 6,
    lock: bool = False,
    **kw,
) -> JsonBlobCache:
    """Open existing database in append mode.

    :path str: Path to the db could be folder or actual file

    :products: Override product dictionary with compatible products loaded from
    the datacube database, this is generally only needed if you intend to add
    datasets to the datacube index directly (i.e. without product matching
    metadata documents).

    :max_db_sz int: Maximum size in bytes database file is allowed to grow to, defaults to 10Gb

    :complevel: Compression level (Zstandard) to use when storing datasets, 1
    fastest, 6 good and still fast, 20+ best but slower.
    """

    subdir = Path(path).is_dir()

    if max_db_sz is None:
        max_db_sz = 10 * (1 << 30)

    db = lmdb.open(
        path,
        subdir=subdir,
        max_dbs=8,
        map_size=max_db_sz,
        lock=lock,
        create=False,
        readonly=False,
        **kw,
    )

    return _from_existing_db(db, complevel=complevel)


def create_cache(
    path: str,
    complevel: int = 6,
    zdict: Optional[bytes] = None,
    max_db_sz: Optional[int] = None,
    truncate: bool = False,
    lock: bool = False,
    subdir: bool = False,
    **kw,
) -> JsonBlobCache:
    """Create new file database or open existing one.

    :path str: Path where to create new database (this will be a directory with 2 files in it)
    :complevel int: Level of compressions to apply per dataset, bigger is slower but better compression.
    :zdict: Optional pre-trained compression dictionary
    :max_db_sz int: Maximum size in bytes (defaults to 10GiB)
    :lock: By default we assume exclusive access to the file, if you expect sharing set ``lock=True``
    :subdir: If set to True treat path as directory

    :truncate bool: If True wipe out any existing database and create new empty one.
    :kw: Passed on to ``lmdb.open(.., **kw)``
      meminit
      writemap
      sync
      metasync
      ...and other

    """

    if truncate:
        maybe_delete_db(path)

    if max_db_sz is None:
        max_db_sz = 10 * (1 << 30)

    db = lmdb.open(
        path,
        max_dbs=8,
        map_size=max_db_sz,
        create=True,
        readonly=False,
        lock=lock,
        subdir=subdir,
        **kw,
    )

    # If db is not empty just call open on it
    if db.stat()["entries"] > 0:
        return _from_existing_db(db, complevel=complevel)
    return _from_empty_db(db, complevel=complevel, zdict=zdict)


def test_key_to_value():

    for k in ("string", 217987, 215781587158712587, ("AAA", 3)):
        bb = key_to_bytes(k)
        assert isinstance(bb, bytes)

    assert key_to_bytes(UUID(bytes=b"0123456789ABCDEF")) == b"0123456789ABCDEF"
    assert key_to_bytes(b"88") == b"88"


def test_create_cache():
    ss = create_cache("tmp.lmdb", truncate=True)
    print(ss)
    del ss
    ss = open_ro("tmp.lmdb")
    print(ss)



import queue
from threading import Thread

import click
import datacube
from odc import dscache
from odc.dscache.tools import (
    bin_dataset_stream,
    dataset_count,
    db_connect,
    dictionary_from_product_list,
    mk_raw2ds,
    ordered_dss,
    raw_dataset_stream,
)
from odc.dscache.tools.tiling import parse_gridspec

EOS = object()


def qmap(proc, q, eos_marker=None):
    """Converts queue to an iterator.

    For every `item` in the `q` that is not `eos_marker`, `yield proc(item)`
    """
    while True:
        item = q.get(block=True)
        try:
            if item is eos_marker:
                break

            yield proc(item)
        finally:
            q.task_done()


@click.command("slurpy")
@click.option("--env", "-E", type=str, help="Datacube environment name")
@click.option(
    "-z",
    "complevel",
    type=int,
    default=6,
    help="Compression setting for zstandard 1-fast, 9+ good but slow",
)
@click.option(
    "--grid",
    type=str,
    help=(
        "Grid spec or name 'crs;pixel_resolution;shape_in_pixels',"
        "albers_au_25, africa_{10|20|30|60}"
    ),
    default=None,
)
@click.option("--year", type=int, help="Only extract datasets for a given year")
@click.argument("output", type=str, nargs=1)
@click.argument("products", type=str, nargs=-1)
def cli(env, grid, year, output, products, complevel):
    """Extract product(s) to an on disk cache.

    Optionally tile datasets into a grid while extracting (see --grid option)
    """
    # pylint: disable=too-many-locals,too-many-statements

    if len(products) == 0:
        click.echo("Have to supply at least one product")
        raise click.Abort()

    dc = datacube.Datacube(env=env)
    all_prods = {p.name: p for p in dc.index.products.get_all()}

    if len(products) == 1 and products[0].lower() in (":all:", "*"):
        click.echo("Will read all products")
        products = list(all_prods)

    for p in products:
        if p not in all_prods:
            click.echo(f"No such product found: {p}")
            raise click.Abort()

    query = {}
    if year is not None:
        query.update(time=f"{year}")

    click.echo("Getting dataset counts")
    counts = {p: dataset_count(dc.index, product=p, **query) for p in products}

    n_total = 0
    for p, c in counts.items():
        click.echo(f"..{p}: {c:8,d}")
        n_total += c

    if n_total == 0:
        click.echo("No datasets found")
        raise click.Abort()

    click.echo("Training compression dictionary")
    zdict = dictionary_from_product_list(
        dc, products, samples_per_product=50, query=query
    )
    click.echo("..done")

    # TODO: check for overwrite
    cache = dscache.create_cache(
        output, zdict=zdict, complevel=complevel, truncate=True
    )

    raw2ds = mk_raw2ds(all_prods)

    def db_task(products, conn, q):
        for p in products:
            if len(query) == 0:
                dss = map(raw2ds, raw_dataset_stream(p, conn))
            else:
                dss = ordered_dss(dc, product=p, **query)

            for ds in dss:
                q.put(ds)
        q.put(EOS)

    conn = db_connect(cfg=env)
    q = queue.Queue(maxsize=10_000)
    db_thread = Thread(target=db_task, args=(products, conn, q))
    db_thread.start()

    dss = qmap(lambda ds: ds, q, eos_marker=EOS)
    dss = cache.tee(dss)

    cells = {}
    group_prefix = ""
    if grid is not None:
        gs = parse_gridspec(grid)
        # TODO for named gridspecs should we use the name as group_prefix?
        group_prefix = f"epsg{gs.crs.epsg:d}"
        cache.add_grid(gs, group_prefix)
        dss = bin_dataset_stream(gs, dss, cells)

    label = f"Processing ({n_total:8,d})"
    with click.progressbar(dss, label=label, length=n_total) as dss:
        for _ in dss:
            pass

    if grid is not None:
        click.echo(f"Total bins: {len(cells):d}")

        with click.progressbar(
            cells.values(), length=len(cells), label="Saving"
        ) as groups:
            for group in groups:
                cache.add_grid_tile(group_prefix, group.idx, group.dss)

    db_thread.join()
    cache.close()


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



from functools import partial

import click
from odc import dscache
from odc.dscache.tools import bin_dataset_stream
from odc.dscache.tools.tiling import (
    bin_by_native_tile,
    extract_native_albers_tile,
    parse_gridspec,
    web_gs,
)


@click.command("dstiler")
@click.option("--native", is_flag=True, help="Use Landsat Path/Row as grouping")
@click.option(
    "--native-albers",
    is_flag=True,
    help="When datasets are in Albers (AU) grid already",
)
@click.option(
    "--web", type=int, help="Use web map tiling regime at supplied zoom level"
)
@click.option(
    "--grid",
    type=str,
    help="Grid spec or name 'crs;pixel_resolution;shape_in_pixels'|albers_au_25",
    default="albers_au_25",
)
@click.argument("dbfile", type=str, nargs=1)
def cli(native, native_albers, web, grid, dbfile):
    """Add spatial grouping to file db.

    Default grid is Australian Albers (EPSG:3577) with 100k by 100k tiles. But
    you can also group by Landsat path/row (--native), or Google's map tiling
    regime (--web zoom_level)

    \b
    Example for custom --grid:
      - rectangular: 'epsg:6933;-10x10;2000x3000'
                      ^crs      ^y  ^x ^ny  ^nx
      - square     : 'epsg:3857;10;10000'
      - named      : albers_au_25
                     africa_10  (20,30,60 are also available)
    """
    cache = dscache.open_rw(dbfile)
    label = f"Processing {dbfile} ({cache.count:,d} datasets)"
    group_prefix = "grid"
    gs = None

    cells = {}
    if native:
        group_prefix = "native"
        binner = partial(bin_by_native_tile, cells=cells)
    elif native_albers:
        group_prefix = "albers"
        binner = lambda dss: bin_by_native_tile(
            dss, cells, native_tile_id=extract_native_albers_tile
        )
    elif web is not None:
        gs = web_gs(web)
        group_prefix = "web_" + str(web)
        binner = lambda dss: bin_dataset_stream(gs, dss, cells)
    else:
        gs = parse_gridspec(grid)
        group_prefix = f"epsg{gs.crs.epsg:d}"
        binner = lambda dss: bin_dataset_stream(gs, dss, cells)

    if gs is not None:
        click.echo(f"Using gridspec: {gs}")
        cache.add_grid(gs, group_prefix)

    with click.progressbar(cache.get_all(), length=cache.count, label=label) as dss:
        for _ in binner(dss):
            pass

    click.echo(f"Total bins: {len(cells):d}")

    with click.progressbar(cells.values(), length=len(cells), label="Saving") as groups:
        for group in groups:
            cache.add_grid_tile(group_prefix, group.idx, group.dss)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



"""These should probably be in datacube library."""
import datetime
from datetime import timedelta
from random import randint
from types import SimpleNamespace
from typing import Iterator, Optional, Set, Tuple
from uuid import UUID
from warnings import warn

import psycopg2
from datacube import Datacube
from datacube.api.query import Query
from datacube.model import Dataset, Range
from datacube.utils.geometry import Geometry
from pandas import Period


def dataset_count(index, **query):
    """Return number of datasets matching a query."""
    return index.datasets.count(**Query(**query).search_terms)


def count_by_year(index, product, min_year=None, max_year=None):
    """
    Return dictionary Int->Int: `year` -> `dataset count for this year`.

    Only non-empty years are reported.
    """
    # TODO: get min/max from datacube properly
    if min_year is None:
        min_year = 1970
    if max_year is None:
        max_year = datetime.datetime.now().year

    year_count = (
        (year, dataset_count(index, product=product, time=str(year)))
        for year in range(min_year, max_year + 1)
    )

    return {year: c for year, c in year_count if c > 0}


def count_by_month(index, product, year):
    """
    Return 12 integer tuple.

    counts for January, February ... December
    """
    return tuple(
        dataset_count(index, product=product, time=f"{year}-{month:02d}")
        for month in range(1, 12 + 1)
    )


def time_range(begin, end, freq="m"):
    """
    Return tuples of datetime objects aligned to boundaries of requested period.

    (month is default).
    """
    tzinfo = begin.tzinfo
    t = Period(begin, freq)

    def to_pydate(t):
        return t.to_pydatetime(warn=False).replace(tzinfo=tzinfo)

    while True:
        t0, t1 = map(to_pydate, (t.start_time, t.end_time))
        if t0 > end:
            break

        yield (max(t0, begin), min(t1, end))
        t += 1


def month_range(
    year: int, month: int, n: int
) -> Tuple[datetime.datetime, datetime.datetime]:
    """
    Construct month aligned time range.

    Return time range covering n months starting from year, month
    month 1..12
    month can also be negative
    2020, -1 === 2019, 12
    """
    if month < 0:
        return month_range(year - 1, 12 + month + 1, n)

    y2 = year
    m2 = month + n
    if m2 > 12:
        m2 -= 12
        y2 += 1
    dt_eps = datetime.timedelta(microseconds=1)

    return (
        datetime.datetime(year=year, month=month, day=1),
        datetime.datetime(year=y2, month=m2, day=1) - dt_eps,
    )


def season_range(year: int, season: str) -> Tuple[datetime.datetime, datetime.datetime]:
    """
    Season is one of djf, mam, jja, son.

    DJF for year X starts in Dec X-1 and ends in Feb X.
    """
    seasons = dict(djf=-1, mam=3, jja=6, son=9)

    start_month = seasons.get(season.lower())
    if start_month is None:
        raise ValueError(f"No such season {season}, valid seasons are: djf,mam,jja,son")
    return month_range(year, start_month, 3)


def chop_query_by_time(q: Query, freq: str = "m") -> Iterator[Query]:
    """
    Split query along time dimension.

    Given a query over longer period of time, chop it up along the time dimension
    into smaller queries each covering a shorter time period (year, month, week or day).
    """
    qq = dict(**q.search_terms)
    time = qq.pop("time", None)
    if time is None:
        raise ValueError("Need time range in the query")

    for (t0, t1) in time_range(time.begin, time.end, freq=freq):
        yield Query(**qq, time=Range(t0, t1))


def ordered_dss(dc: Datacube, freq: str = "m", key=None, **query):
    """
    Emulate "order by time" streaming interface for datacube queries.

        Basic idea is to perform a lot of smaller queries (shorter time
        periods), sort results then yield them to the calling code.

    :param dc: Datacube instance

    :param freq: 'm' month sized chunks, 'w' week sized chunks, 'd' day

    :param key: Optional sorting function Dataset -> Comparable, for example
                ``lambda ds: (ds.center_time, ds.metadata.region_code)``
    """
    qq = Query(**query)
    if key is None:
        key = lambda ds: ds.center_time

    _last_uuids: Set[UUID] = set()

    for q in chop_query_by_time(qq, freq=freq):
        _dss = dc.find_datasets(**q.search_terms)
        dss = [ds for ds in _dss if ds.id not in _last_uuids]
        _last_uuids = {ds.id for ds in _dss}
        dss.sort(key=key)
        yield from dss


def chopped_dss(dc: Datacube, freq: str = "m", **query):
    """
    Emulate streaming interface for datacube queries.

    Basic idea is to perform a lot of smaller queries (shorter time
    periods)
    """
    qq = Query(**query)
    _last_uuids: Set[UUID] = set()

    for q in chop_query_by_time(qq, freq=freq):
        _dss = dc.find_datasets(**q.search_terms)
        dss = [ds for ds in _dss if ds.id not in _last_uuids]
        _last_uuids = {ds.id for ds in _dss}
        yield from dss


def bin_dataset_stream(gridspec, dss, cells, persist=None):
    """
    Intersect Grid Spec cells with Datasets.

    :param gridspec: GridSpec
    :param dss: Sequence of datasets (can be lazy)
    :param cells: Dictionary to populate with tiles
    :param persist: Dataset -> SomeThing mapping, defaults to keeping dataset id only

    The ``cells`` dictionary is a mapping from (x,y) tile index to object with the
    following properties:

     .idx     - tile index (x,y)
     .geobox  - tile geobox
     .utc_offset - timedelta to add to timestamp to get day component in local time
     .dss     - list of UUIDs, or results of ``persist(dataset)`` if custom ``persist`` is supplied
    """
    geobox_cache = {}

    def default_persist(ds):
        return ds.id

    def register(tile, geobox, val):
        cell = cells.get(tile)
        if cell is None:
            utc_ofset = solar_offset(geobox.extent)
            cells[tile] = SimpleNamespace(
                geobox=geobox, idx=tile, utc_offset=utc_ofset, dss=[val]
            )
        else:
            cell.dss.append(val)

    if persist is None:
        persist = default_persist

    for ds in dss:
        ds_val = persist(ds)

        if ds.extent is None:
            warn(f"Dataset without extent info: {str(ds.id)}")
            continue

        for tile, geobox in gridspec.tiles_from_geopolygon(
            ds.extent, geobox_cache=geobox_cache
        ):
            register(tile, geobox, ds_val)

        yield ds


def bin_dataset_stream2(gridspec, dss, geobox_cache=None):
    """
    For every input dataset compute tiles of the GridSpec it overlaps with.

    Iterable[Dataset] -> Iterator[(Dataset, List[Tuple[int, int]])]
    """
    if geobox_cache is None:
        geobox_cache = {}

    for ds in dss:
        if ds.extent is None:
            warn(f"Dataset without extent info: {ds.id}")
            tiles = []
        else:
            tiles = [
                tile
                for tile, _ in gridspec.tiles_from_geopolygon(
                    ds.extent, geobox_cache=geobox_cache
                )
            ]

        yield ds, tiles


def all_datasets(
    dc: Datacube, product: str, read_chunk: int = 1000, limit: Optional[int] = None
):
    """
    Properly lazy version of ``dc.find_datasets_lazy(product=product)``.

    Uses db cursors to reduce latency of results arriving from the database, original
    method in datacube fetches entire query result from the database and only then lazily
    constructs Dataset objects from the returned table.

    :param dc: Datacube object
    :param product: Product name to extract
    :param read_chunk: Number of datasets per page
    :param limit: Optionally constrain query to just return that many dataset
    """
    assert isinstance(limit, (int, type(None)))

    db = psycopg2.connect(str(dc.index.url))
    _limit = "" if limit is None else f"LIMIT {limit}"

    _product = dc.index.products.get_by_name(product)
    if _product is None:
        raise ValueError(f"No such product: {product}")

    query = f"""select
jsonb_build_object(
  'product', %(product)s,
  'uris', array((select _loc_.uri_scheme ||':'||_loc_.uri_body
                 from agdc.dataset_location as _loc_
                 where _loc_.dataset_ref = agdc.dataset.id and _loc_.archived is null
                 order by _loc_.added desc, _loc_.id desc)),
  'metadata', metadata) as dataset
from agdc.dataset
where archived is null
and dataset_type_ref = (select id from agdc.dataset_type where name = %(product)s)
{_limit};
"""
    cursor_name = f"c{randint(0, 0xFFFF):04X}"
    with db.cursor(name=cursor_name) as cursor:
        cursor.execute(query, dict(product=product))

        while True:
            chunk = cursor.fetchmany(read_chunk)
            if not chunk:
                break
            for (ds,) in chunk:
                yield Dataset(_product, ds["metadata"], ds["uris"])


def mid_longitude(geom: Geometry) -> float:
    """Return longitude of the middle point of the geomtry."""
    ((lon,), _) = geom.centroid.to_crs("epsg:4326").xy
    return lon


def solar_offset(geom: Geometry, precision: str = "h") -> timedelta:
    """
    Given a geometry compute offset to add to UTC timestamp to get solar day right.

    This only work when geometry is "local enough".
    :param precision: one of ``'h'`` or ``'s'``, defaults to hour precision
    """
    lon = mid_longitude(geom)

    if precision == "h":
        return timedelta(hours=int(lon * 24 / 360 + 0.5))

    # 240 == (24*60*60)/360 (seconds of a day per degree of longitude)
    return timedelta(seconds=int(lon * 240))



from pathlib import Path
from typing import (
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)
from uuid import UUID

import toolz
from datacube.model import (
    Dataset,
    DatasetType,
    GridSpec,
    MetadataType,
    metadata_from_doc,
)
from datacube.utils.geometry import CRS

from . import _jsoncache as base
from ._text import split_and_check

# pylint: disable=invalid-name,too-many-public-methods

ProductCollection = Union[
    Iterator[DatasetType], List[DatasetType], Dict[str, DatasetType]
]
Document = base.Document
LaxUUID = base.LaxUUID
TileIdx = Union[Tuple[int, int], Tuple[str, int, int]]


def ds2doc(ds) -> Tuple[UUID, Document]:
    return (ds.id, dict(uris=ds.uris, product=ds.type.name, metadata=ds.metadata_doc))


def doc2ds(
    doc: Optional[Document], products: Dict[str, DatasetType]
) -> Optional[Dataset]:
    if doc is None:
        return None

    p = products.get(doc["product"], None)
    if p is None:
        raise ValueError(f"No product named: {doc['product']}")
    return Dataset(p, doc["metadata"], uris=doc["uris"])


def gs2doc(gs: GridSpec) -> base.Document:
    return dict(
        crs=str(gs.crs),
        tile_size=list(gs.tile_size),
        resolution=list(gs.resolution),
        origin=list(gs.origin),
    )


def doc2gs(doc: Document) -> GridSpec:
    return GridSpec(
        crs=CRS(doc["crs"]),
        tile_size=tuple(doc["tile_size"]),  # type: ignore
        resolution=tuple(doc["resolution"]),  # type: ignore
        origin=tuple(doc["origin"]),  # type: ignore
    )


def build_dc_product_map(
    metadata_json: Document, products_json: Document
) -> Tuple[Dict[str, MetadataType], Dict[str, DatasetType]]:
    mm = toolz.valmap(metadata_from_doc, metadata_json)

    def mk_product(doc, name):
        mt = doc.get("metadata_type")
        if mt is None:
            raise ValueError("Missing metadata_type key in product definition")
        metadata = mm.get(mt)

        if metadata is None:
            raise ValueError(f"No such metadata {mt} for product {name}")

        return DatasetType(metadata, doc)

    return mm, {k: mk_product(doc, k) for k, doc in products_json.items()}


def _metadata_from_products(
    products: Dict[str, DatasetType]
) -> Dict[str, MetadataType]:
    mm = {}
    for p in products.values():
        m = p.metadata_type
        if m.name not in mm:
            mm[m.name] = m

    return mm


def mk_group_name(idx: TileIdx, name: str = "unnamed_grid") -> str:
    x, y = idx[-2:]
    if len(idx) == 2:
        return f"{name}/{x:+05d}/{y:+05d}"
    if len(idx) == 3:
        t = idx[0]
        return f"{name}/{t}/{x:+05d}/{y:+05d}"
    raise ValueError("Expect index in (x, y) or (t, x, y) format")


def parse_group_name(group_name: str) -> Tuple[TileIdx, str]:
    """Return an
      ((int, int), prefix:str)       x,y
      ((str, int, int), prefix:str)  t,x,y

    tuple from group name.

    Expects group to be in the form
     {prefix}/{x}/{y}
    or
     {prefix}/{t}/{x}/{y}

    raises ValueError if group_name is not in the expected format.
    """

    try:
        prefix, *tidx = split_and_check(group_name, "/", (3, 4))
        x, y = map(int, tidx[-2:])
    except ValueError:
        raise ValueError("Bad group name: " + group_name) from None

    if len(tidx) == 2:
        return (x, y), prefix

    return (cast(str, tidx[0]), x, y), prefix


class DatasetCache:
    """
    info:
       version: 4-bytes
       zdict: pre-trained compression dictionary, optional
       product/{name}: json
       metadata/{name}: json
       grid/{name}: json

    groups:
       Each group is named list of uuids

    udata:
       arbitrary user data (TODO)

    ds:
       uuid: compressed(json({product: str,
                              uris: [str],
                              metadata: object}))
    """

    def __init__(
        self, db: base.JsonBlobCache, products: Optional[ProductCollection] = None
    ):
        """Don't use this directly, use create_cache or open_(rw|ro)."""

        if products is None:
            metadata, products = build_dc_product_map(
                db.get_info_dict("metadata/"), db.get_info_dict("product/")
            )
        else:
            if not isinstance(products, dict):
                products = {p.name: p for p in products}

            metadata = _metadata_from_products(products)

        self._db = db
        self._products = products
        self._metadata = metadata

    def close(self):
        """Write any outstanding product/metadata definitions (if in write mode) and
        close database file.

        """
        self._db.close()

    @staticmethod
    def train_dictionary(dss: Iterator[Dataset], dict_sz=8 * 1024) -> Optional[bytes]:
        """Given a finite sequence of Datasets train zstandard compression dictionary of a given size.

        Accepts both `Dataset` as well as "raw" datasets.

        Will return None if input sequence is empty.
        """
        docs = list(map(ds2doc, dss))
        return base.JsonBlobCache.train_dictionary(docs, dict_sz=dict_sz)

    @property
    def readonly(self) -> bool:
        return self._db.readonly

    @property
    def count(self) -> int:
        return self._db.count

    def put_group(self, name: str, uuids: Collection[UUID]):
        """Group is a named list of uuids"""
        self._db.put_group(name, uuids)

    def get_group(self, name: str) -> Optional[List[UUID]]:
        """Group is a named list of uuids"""
        return self._db.get_group(name)

    def groups(self, raw=False, prefix=None):
        """Get list of tuples (group_name, group_size).

        :raw bool: Normally names are returned as strings, supplying raw=True
        would return bytes instead, this is needed if you are using group names
        that are not strings, like integers or tuples of basic types.

        :prefix str|bytes: Only report groups with name starting with prefix
        """
        return self._db.groups(raw=raw, prefix=prefix)

    @property
    def products(self) -> Dict[str, DatasetType]:
        return self._products

    @property
    def metadata(self) -> Dict[str, MetadataType]:
        return self._metadata

    def _add_metadata(self, metadata: MetadataType, transaction: base.MaybeTransaction):
        self._metadata[metadata.name] = metadata
        self._db.append_info_dict(
            "metadata/", {metadata.name: metadata.definition}, transaction
        )

    def _add_product(self, product: DatasetType, transaction: base.MaybeTransaction):
        if product.metadata_type.name not in self._metadata:
            self._add_metadata(product.metadata_type, transaction)

        self._products[product.name] = product
        self._db.append_info_dict(
            "product/", {product.name: product.definition}, transaction
        )

    def _ds2doc(self, ds: Dataset) -> Tuple[UUID, Document]:
        if ds.type.name not in self._products:
            self._add_product(ds.type, self._db.current_transaction)
        return ds2doc(ds)

    def bulk_save(self, dss: Iterable[Dataset]):
        docs = (self._ds2doc(ds) for ds in dss)
        return self._db.bulk_save(docs)

    def tee(
        self, dss: Iterable[Dataset], max_transaction_size: int = 10000
    ) -> Iterator[Dataset]:
        """Given a lazy stream of datasets persist them to disk and then pass through
        for further processing.
        :dss: stream of datasets
        :max_transaction_size int: How often to commit results to disk
        """
        return self._db.tee(
            dss, max_transaction_size=max_transaction_size, transform=self._ds2doc
        )

    def get(self, uuid: LaxUUID) -> Optional[Dataset]:
        """Extract single dataset with a given uuid, or return None if not found"""
        return doc2ds(self._db.get(uuid), self._products)

    def get_all(self) -> Iterator[Dataset]:
        for _, v in self._db.get_all():
            ds = doc2ds(v, self._products)
            assert ds is not None
            yield ds

    def stream_group(self, group_name: str) -> Iterator[Dataset]:
        for _, v in self._db.stream_group(group_name):
            ds = doc2ds(v, self._products)
            assert ds is not None
            yield ds

    @property
    def grids(self) -> Dict[str, GridSpec]:
        """Grids defined for this dataset cache"""
        return {
            key: doc2gs(value) for key, value in self._db.get_info_dict("grid/").items()
        }

    def add_grid(self, gs: GridSpec, name: str):
        """Register a grid"""
        self._db.append_info_dict("grid/", {name: gs2doc(gs)})

    def add_grid_tile(self, grid: str, idx: TileIdx, uuids: Collection[UUID]):
        """Add list of dataset UUIDs to a tile"""
        key = mk_group_name(idx, grid)
        self._db.put_group(key, uuids)

    def add_grid_tiles(self, grid: str, tiles: Dict[TileIdx, List[UUID]]):
        """Add multiple tiles to a grid"""
        for idx, uuids in tiles.items():
            self.add_grid_tile(grid, idx, uuids)

    def tiles(self, grid: str) -> List[Tuple[TileIdx, int]]:
        """Return tile indexes and dataset counts"""

        def tile_index(group_name):
            idx, prefix = parse_group_name(group_name)
            assert prefix == grid
            return idx

        return [
            (tile_index(group_name), count)
            for group_name, count in self.groups(prefix=grid + "/")
        ]

    def stream_grid_tile(self, idx: TileIdx, grid: str) -> Iterator[Dataset]:
        """Iterate over datasets in a given tile"""
        return self.stream_group(mk_group_name(idx, grid))

    def append_info_dict(
        self, prefix: str, oo: Document, transaction: base.MaybeTransaction = None
    ):
        self._db.append_info_dict(prefix, oo, transaction=transaction)

    def get_info_dict(
        self, prefix: str, transaction: base.MaybeTransaction = None
    ) -> Document:
        return self._db.get_info_dict(prefix, transaction=transaction)

    def clear_info_dict(self, prefix: str, transaction: base.MaybeTransaction = None):
        return self._db.clear_info_dict(prefix, transaction=transaction)

    @property
    def path(self) -> Path:
        return self._db.path

    @staticmethod
    def open_ro(path: str, lock: bool = False) -> "DatasetCache":
        return open_ro(path, lock=lock)

    @staticmethod
    def open_rw(path: str, **kw) -> "DatasetCache":
        return open_rw(path, **kw)

    @staticmethod
    def create(
        path: str,
        complevel: int = 6,
        zdict: Optional[bytes] = None,
        max_db_sz: Optional[int] = None,
        lock: bool = False,
        subdir: bool = False,
        truncate: bool = False,
        **kw,
    ) -> "DatasetCache":
        return create_cache(
            path,
            complevel=complevel,
            zdict=zdict,
            max_db_sz=max_db_sz,
            lock=lock,
            subdir=subdir,
            truncate=truncate,
            **kw,
        )

    @staticmethod
    def exists(path: str) -> bool:
        return base.db_exists(path)


def open_ro(
    path: str, products: Optional[ProductCollection] = None, lock: bool = False, **kw
) -> DatasetCache:
    """Open existing database in readonly mode.

    .. note::

    default mode assumes db file is static (not being modified
    externally), if this is not the case, supply `lock=True` parameter.

    :path str: Path to the db could be folder or actual file

    :products: Override product dictionary with compatible products loaded from
    the datacube database, this is generally only needed if you intend to add
    datasets to the datacube index directly (i.e. without product matching
    metadata documents).

    :lock bool: Supply True if external process is changing DB concurrently.

    """

    db = base.open_ro(path, lock=lock, **kw)
    return DatasetCache(db, products=products)


def open_rw(
    path: str,
    products: Optional[ProductCollection] = None,
    max_db_sz: Optional[int] = None,
    complevel: int = 6,
    lock: bool = False,
    **kw,
) -> DatasetCache:
    """Open existing database in append mode.

    :path str: Path to the db could be folder or actual file

    :products: Override product dictionary with compatible products loaded from
    the datacube database, this is generally only needed if you intend to add
    datasets to the datacube index directly (i.e. without product matching
    metadata documents).

    :max_db_sz int: Maximum size in bytes database file is allowed to grow to, defaults to 10Gb

    :complevel: Compression level (Zstandard) to use when storing datasets, 1
    fastest, 6 good and still fast, 20+ best but slower.
    """
    db = base.open_rw(path, max_db_sz=max_db_sz, complevel=complevel, lock=lock, **kw)
    return DatasetCache(db, products)


def create_cache(
    path: str,
    complevel: int = 6,
    zdict: Optional[bytes] = None,
    max_db_sz: Optional[int] = None,
    truncate: bool = False,
    lock: bool = False,
    subdir: bool = False,
    **kw,
) -> DatasetCache:
    """Create new file database or open existing one.

    :path str: Path where to create new database (this will be a directory with 2 files in it)
    :complevel int: Level of compressions to apply per dataset, bigger is slower but better compression.
    :zdict: Optional pre-trained compression dictionary
    :max_db_sz int: Maximum size in bytes (defaults to 10GiB)
    :lock: By default we assume exclusive access to the file, if you expect sharing set ``lock=True``
    :subdir: If set to True treat path as directory

    :truncate bool: If True wipe out any existing database and create new empty one.
    :kw: Passed on to ``lmdb.open(.., **kw)``
      meminit
      writemap
      sync
      metasync
      ...and other

    """
    db = base.create_cache(
        path,
        complevel=complevel,
        zdict=zdict,
        max_db_sz=max_db_sz,
        truncate=truncate,
        lock=lock,
        subdir=subdir,
        **kw,
    )
    return DatasetCache(db)



from math import floor, pi
from types import SimpleNamespace
from typing import Dict, Optional, Tuple

import toolz
from datacube.model import Dataset, GridSpec
from datacube.utils.geometry import CRS

from .._text import parse_range_int, split_and_check

epsg3577 = CRS("epsg:3577")
epsg6933 = CRS("epsg:6933")

# Origin was chosen such that there are no negative indexed tiles for any valid
# point within a given CRS's valid region, and also making sure that x=0,y=0
# lines fall on tile edges.
#
#  au = gridspec_from_crs(CRS("epsg:3577"),
#                         tile_size=(96_000, 96_000),
#                         pad_yx=(5, 5))
#
#  So AU tiles with index `y < 5 or x < 5` are outside of the valid range of EPSG:3577.
#
GRIDS = {
    "albers_au_25": GridSpec(
        crs=epsg3577, tile_size=(100_000.0, 100_000.0), resolution=(-25, 25)
    ),
    "au": GridSpec(
        crs=epsg3577,
        tile_size=(96_000.0, 96_000.0),
        resolution=(-96_000, 96_000),
        origin=(-5472000.0, -2688000.0),
    ),
    **{
        f"au_{n}": GridSpec(
            crs=epsg3577,
            tile_size=(96_000.0, 96_000.0),
            resolution=(-n, n),
            origin=(-5472000.0, -2688000.0),
        )
        for n in (10, 20, 30, 60)
    },
    "global": GridSpec(
        crs=epsg6933,
        tile_size=(96_000.0, 96_000.0),
        resolution=(-96_000, 96_000),
        origin=(-7392000, -17376000),
    ),
    **{
        f"global_{n}": GridSpec(
            crs=epsg6933,
            tile_size=(96_000.0, 96_000.0),
            resolution=(-n, n),
            origin=(-7392000, -17376000),
        )
        for n in (10, 20, 30, 60)
    },
}


# Inject aliases for Africa
GRIDS["africa"] = GRIDS["global"]
for r in (10, 20, 30, 60):
    GRIDS[f"africa_{r}"] = GRIDS[f"global_{r}"]


def web_gs(zoom: int, tile_size: int = 256) -> GridSpec:
    """
    Construct grid spec compatible with TerriaJS requests at a given level.

    Tile indexes should be the same as google maps, except that Y component is negative,
    this is a limitation of GridSpec class, you can not have tile index direction be
    different from axis direction, but this is what google indexing is using.

    http://www.maptiler.org/google-maps-coordinates-tile-bounds-projection/
    """
    R = 6378137

    origin = pi * R
    res0 = 2 * pi * R / tile_size
    res = res0 * (2 ** (-zoom))
    tsz = 2 * pi * R * (2 ** (-zoom))  # res*tile_size

    return GridSpec(
        crs=CRS("epsg:3857"),
        tile_size=(tsz, tsz),
        resolution=(-res, res),  # Y,X
        origin=(origin - tsz, -origin),
    )  # Y,X


def extract_native_albers_tile(
    ds: Dataset, tile_size: float = 100_000.0
) -> Tuple[int, int]:
    ll = toolz.get_in(
        "grid_spatial.projection.geo_ref_points.ll".split("."), ds.metadata_doc
    )
    assert ll is not None
    return (int(ll["x"] / tile_size), int(ll["y"] / tile_size))


def extract_ls_path_row(ds: Dataset) -> Optional[Tuple[int, int]]:
    full_id = ds.metadata_doc.get("tile_id")

    if full_id is None:
        full_id = toolz.get_in(["usgs", "scene_id"], ds.metadata_doc)

    if full_id is None:
        return None
    return (int(full_id[3:6]), int(full_id[6:9]))


def bin_by_native_tile(dss, cells, persist=None, native_tile_id=None):
    """Group datasets by native tiling, like path/row for Landsat.

    :param dss: Sequence of datasets (can be lazy)
    :param cells: Dictionary to populate with tiles

    :param persist: Dataset -> SomeThing mapping, defaults to keeping dataset
    id only

    :param native_tile_id: Dataset -> Key, defaults to extracting `path,row`
    tuple from metadata's `tile_id` field, but could be anything, only
    constraint is that Key value can be used as index to python dict.
    """

    def default_persist(ds):
        return ds.id

    def register(tile, val):
        cell = cells.get(tile)
        if cell is None:
            cells[tile] = SimpleNamespace(idx=tile, dss=[val])
        else:
            cell.dss.append(val)

    native_tile_id = native_tile_id or extract_ls_path_row
    persist = persist or default_persist

    for ds in dss:
        ds_val = persist(ds)
        tile = native_tile_id(ds)
        if tile is None:
            raise ValueError("Missing tile id")

        register(tile, ds_val)
        yield ds


def _parse_gridspec_string(s: str) -> GridSpec:
    """
    "epsg:6936;10;9600"
    "epsg:6936;-10x10;9600x9600"
    """

    crs, _res, _shape = split_and_check(s, ";", 3)
    try:
        if "x" in _res:
            r1, r2 = tuple(float(v) for v in split_and_check(_res, "x", 2))
            res = (r1, r2)
        else:
            tmp = float(_res)
            res = (-tmp, tmp)

        if "x" in _shape:
            shape = parse_range_int(_shape, separator="x")
        else:
            sz = int(_shape)
            shape = (sz, sz)
    except ValueError:
        raise ValueError(f"Failed to parse gridspec: {s}") from None

    t1, t2 = tuple(abs(n * res) for n, res in zip(res, shape))
    tsz = (t1, t2)

    return GridSpec(crs=CRS(crs), tile_size=tsz, resolution=res, origin=(0, 0))


def _norm_gridspec_name(s: str) -> str:
    return s.replace("-", "_")


def parse_gridspec(s: str, grids: Optional[Dict[str, GridSpec]] = None) -> GridSpec:
    """
    "africa_10"
    "epsg:6936;10;9600"
    "epsg:6936;-10x10;9600x9600"
    """
    if grids is None:
        grids = GRIDS

    named_gs = grids.get(_norm_gridspec_name(s))
    if named_gs is not None:
        return named_gs

    return _parse_gridspec_string(s)


def parse_gridspec_with_name(
    s: str, grids: Optional[Dict[str, GridSpec]] = None
) -> Tuple[str, GridSpec]:
    if grids is None:
        grids = GRIDS

    named_gs = grids.get(_norm_gridspec_name(s))
    if named_gs is not None:
        return (s, named_gs)

    gs = _parse_gridspec_string(s)
    s = s.replace(";", "_")
    return (s, gs)


def gridspec_from_crs(
    crs: CRS,
    tile_size: Tuple[float, float] = (96_000, 96_000),
    pad_yx: Tuple[int, int] = (0, 0),
    resolution: Optional[Tuple[float, float]] = None,
):
    """
    Compute GridSpec such that there are no negative tiles overlapping with the
    valid region of the target CRS.

    :param crs: Coordinate System used to define the grid
    :param tile_size: (Y, X) size of each tile, in CRS units
    :param pad_yx: (Y, X) safety margin in number of tiles
    :param resolution: (Y, X) size of each data point in the grid, in CRS units. Y will
                       usually be negative.
    """
    if resolution is None:
        resolution = (-tile_size[0], tile_size[1])

    valid_region = crs.valid_region
    assert valid_region is not None

    x0, y0, _, _ = valid_region.to_crs(crs).boundingbox
    # index of the tile containing bottom left corner
    iy, ix = (int(floor(v / tsz)) for v, tsz in zip((y0, x0), tile_size))

    y0_, x0_ = (
        float((idx - pad) * tsz) for (idx, pad, tsz) in zip((iy, ix), pad_yx, tile_size)
    )

    return GridSpec(crs, tile_size, resolution=resolution, origin=(y0_, x0_))



"""
Tools for dealing with datacube db
"""
import random
from typing import Any, Dict, Optional, Tuple

import datacube.utils.geometry as geom
import psycopg2
from datacube import Datacube
from datacube.api.grid_workflow import Tile
from datacube.config import LocalConfig
from datacube.model import Dataset, GridSpec

from .. import DatasetCache, TileIdx
from ._index import (
    all_datasets,
    bin_dataset_stream,
    chopped_dss,
    dataset_count,
    ordered_dss,
    solar_offset,
)

__all__ = (
    "DcTileExtract",
    "all_datasets",
    "bin_dataset_stream",
    "dataset_count",
    "db_connect",
    "dictionary_from_product_list",
    "chopped_dss",
    "grid_tiles_to_geojson",
    "gs_albers",
    "mk_raw2ds",
    "ordered_dss",
    "raw_dataset_stream",
    "solar_offset",
)


def dictionary_from_product_list(
    dc, products, samples_per_product=10, dict_sz=8 * 1024, query=None
):

    """Get a sample of datasets from a bunch of products and train compression
    dictionary.

    dc -- Datcube object
    products -- list of product names
    samples_per_product -- number of datasets per product to use for training
    dict_sz -- size of dictionary in bytes
    """

    if isinstance(products, str):
        products = [products]

    if query is None:
        query = {}

    limit = samples_per_product * 10

    samples = []
    for p in products:
        dss = dc.find_datasets(product=p, limit=limit, **query)
        random.shuffle(dss)
        samples.extend(dss[:samples_per_product])

    if len(samples) == 0:
        return None

    return DatasetCache.train_dictionary(samples, dict_sz)  # type: ignore


def db_connect(cfg=None):
    """Create database connection from datacube config.

    cfg:
      None -- use default datacube config
      str  -- use config with a given name

      LocalConfig -- use loaded config object
    """
    if isinstance(cfg, str) or cfg is None:
        cfg = LocalConfig.find(env=cfg)

    cfg_remap = dict(
        dbname="db_database",
        user="db_username",
        password="db_password",
        host="db_hostname",
        port="db_port",
    )

    pg_cfg = {k: cfg.get(cfg_name, None) for k, cfg_name in cfg_remap.items()}

    return psycopg2.connect(**pg_cfg)


def mk_raw2ds(products):
    """Convert "raw" dataset to `datacube.model.Dataset`.

    products -- dictionary from product name to `Product` object (`DatasetType`)

    returns: function that maps: `dict` -> `datacube.model.Dataset`

    This function can raise `ValueError` if dataset product is not found in the
    supplied products dictionary.


    Here "raw dataset" is just a python dictionary with fields:

    - product: str -- product name
    - uris: [str] -- list of dataset uris
    - metadata: dict -- dataset metadata document

    see `raw_dataset_stream`

    """

    def raw2ds(ds):
        product = products.get(ds["product"], None)
        if product is None:
            raise ValueError(f"Missing product: {ds['product']}")
        return Dataset(product, ds["metadata"], uris=ds["uris"])

    return raw2ds


def raw_dataset_stream(product, db, read_chunk=100, limit=None):
    """Given a product name stream all "active" datasets from db that belong to that product.

    Datasets are returned in "raw form", basically just a python dictionary with fields:

    - product: str -- product name
    - uris: [str] -- list of dataset uris
    - metadata: dict -- dataset metadata document
    """

    assert isinstance(limit, (int, type(None)))

    if isinstance(db, str) or db is None:
        db = db_connect(db)

    _limit = f"LIMIT {limit:d}" if limit else ""
    query = f"""
select
jsonb_build_object(
  'product', %(product)s,
  'uris', array((select _loc_.uri_scheme ||':'||_loc_.uri_body
                 from agdc.dataset_location as _loc_
                 where _loc_.dataset_ref = agdc.dataset.id and _loc_.archived is null
                 order by _loc_.added desc, _loc_.id desc)),
  'metadata', metadata) as dataset
from agdc.dataset
where archived is null
and dataset_type_ref = (select id from agdc.dataset_type where name = %(product)s)
{_limit};
"""

    cur = db.cursor(name=f"c{random.randint(0, 0xFFFF):04X}")
    cur.execute(query, dict(product=product))

    while True:
        chunk = cur.fetchmany(read_chunk)
        if not chunk:
            break

        for (ds,) in chunk:
            yield ds

    cur.close()


def gs_albers():
    return GridSpec(
        crs=geom.CRS("EPSG:3577"), tile_size=(100000.0, 100000.0), resolution=(-25, 25)
    )


# pylint: disable=too-few-public-methods
class DcTileExtract:
    """Construct ``datacube.api.grid_workflow.Tile`` object from dataset cache."""

    def __init__(self, cache, grid=None, group_by="time"):

        gs = cache.grids.get(grid, None)
        if gs is None:
            raise ValueError(f"No such grid: ${grid}")

        self._cache = cache
        self._grid = grid
        self._gs = gs
        self._default_groupby = group_by

    def __call__(self, tile_idx, _y=None, group_by=None):
        if _y is not None:
            tile_idx = (tile_idx, _y)

        if group_by is None:
            group_by = self._default_groupby

        dss = list(self._cache.stream_grid_tile(tile_idx, grid=self._grid))
        sources = Datacube.group_datasets(dss, group_by)

        geobox = self._gs.tile_geobox(tile_idx)
        return Tile(sources, geobox)


def grid_tiles_to_geojson(
    cache: DatasetCache,
    grid: str,
    style: Optional[Dict[str, Any]] = None,
    wrapdateline: bool = False,
) -> Dict[str, Any]:
    """
    Render tiles of a given grid to GeoJSON.

    each tile is a GeoJSON Feature with following properties:

     .title   -- str: grid index as a string
     .count   -- int: count of datasets overlapping with this grid tile

    :param cache: Dataset cache from which to read gridspec and tiles
    :param grid: Name of the grid to dump
    :param style: Optional style dictionary (will be included in every tile)
    :param wrapdateline: If set check for lon=180 intersect and do "the right thing" (slower)
    """
    if style is None:
        # these are understood by github renderer
        style = {"fill-opacity": 0, "stroke-width": 0.5}

    gs = cache.grids.get(grid, None)
    if gs is None:
        raise ValueError(f"No such grid: {grid}")

    resolution = abs(gs.tile_size[0]) / 4  # up to 4 points per side

    def mk_feature(tidx: TileIdx, count: int) -> Dict[str, Any]:
        if len(tidx) == 3:
            _xy: Tuple[int, int] = tidx[1:]  # type: ignore
        else:
            _xy: Tuple[int, int] = tidx  # type: ignore

        return dict(
            type="Feature",
            geometry=gs.tile_geobox(_xy)
            .extent.to_crs(
                "epsg:4326", resolution=resolution, wrapdateline=wrapdateline
            )
            .json,
            properties={
                "title": f"{_xy[0]:+05d},{_xy[1]:+05d}",
                "count": count,
                **style,  # type: ignore
            },
        )

    features = [mk_feature(tidx, cc) for tidx, cc in cache.tiles(grid)]

    return {"type": "FeatureCollection", "features": features}



"""ODC Dataset Cache"""
from ._version import __version__  # isort:skip  this has to be 1st import
from ._dscache import DatasetCache, TileIdx, create_cache, open_ro, open_rw
from ._jsoncache import JsonBlobCache, db_exists

__all__ = (
    "create_cache",
    "open_ro",
    "open_rw",
    "db_exists",
    "TileIdx",
    "DatasetCache",
    "JsonBlobCache",
    "__version__",
)



from typing import Tuple, Union


def split_and_check(
    s: str, separator: str, n: Union[int, Tuple[int, ...]]
) -> Tuple[str, ...]:
    """Turn string into tuple, checking that there are exactly as many parts as expected.
    :param s: String to parse
    :param separator: Separator character
    :param n: Expected number of parts, can be a single integer value or several,
              example `(2, 3)` accepts 2 or 3 parts.
    """
    if isinstance(n, int):
        n = (n,)

    parts = s.split(separator)
    if len(parts) not in n:
        raise ValueError(f'Failed to parse "{s}"')
    return tuple(parts)


def parse_range_int(s: str, separator: str = ":") -> Tuple[int, int]:
    """Parse str(<int>:<int>) -> (int, int)"""
    try:
        _in, _out = (int(x) for x in split_and_check(s, separator, 2))
    except ValueError:
        raise ValueError(f'Expect <int>{separator}<int> syntax, got "{s}"') from None

    return (_in, _out)



import timeit
from types import SimpleNamespace


def _rr2s(r):
    return """
Count: {r.count:,d}
       {fps:.1f} per second
Total: {r.total:6.3f} sec
TTFB : {r.ttfb:6.3f} sec
.....: {r.uu:032X}
..
""".format(
        r=r, fps=r.count / r.total
    ).strip()


def ds_stream_test_func(dss, get_uuid=None):
    def default_get_uuid(ds):
        return ds.id

    if get_uuid is None:
        get_uuid = default_get_uuid

    timer = timeit.default_timer
    uu = 0
    count = 0
    t00 = timer()
    t0 = None
    for ds in dss:
        t0 = t0 or timer()
        uu ^= get_uuid(ds).int
        count += 1

    t_end = timer()

    rr = SimpleNamespace(uu=uu, count=count, ttfb=t0 - t00, total=t_end - t00, text="")
    rr.text = _rr2s(rr)
    return rr
